"""Define functions for the interface with the user."""

from __future__ import annotations

import os
import re

from typing import Any, Optional, Type, Union, cast

import inquirer

from colorama import Fore, Style, init
from jinja2 import Environment

from scicookie.logs import SciCookieErrorType, SciCookieLogs

ENV = Environment(
    autoescape=False,
    variable_start_string="${{",
    variable_end_string="}}",
)


# Initialize Colorama
init()


def _create_question(
    question_id: str, question: dict
) -> Optional[inquirer.questions.Question]:
    # config required
    question_type = question.get("type", "")
    # todo: implement help text int he prompt
    # question_help = question.get("help")
    # todo: implement depends on workflow, it needs refactoring the code
    #       because it needs access to the answer
    # question_depends_on = question.get("depends_on")
    # it should allow multiple conditionals

    question_types_available = [
        "text",
        "single-choice",
        "multiple-choices",
        "confirm",
    ]

    if question_type not in question_types_available:
        SciCookieLogs.raise_error(
            f"The question type ({question_type}) is not available ("
            f"{question_types_available}).",
            SciCookieErrorType.SCICOOKIE_INVALID_CONFIGURATION,
        )

    content = {"message": ""}

    if question.get("choices"):
        content["choices"] = question.get("choices", [])

    fn_questions: dict[str, Type[inquirer.questions.Question]] = {
        "text": inquirer.Text,
        "single-choice": inquirer.List,
        "multiple-choices": inquirer.Checkbox,
        "confirm": inquirer.Confirm,
    }

    return fn_questions[question_type](
        question_id,
        **content,
    )


def check_visibility(
    question: dict[str, Any], answers: dict[str, str]
) -> bool:
    """Check if dependencies are satisfied."""
    # validation
    is_visible: Union[str, bool] = question.get("visible", False)

    if isinstance(is_visible, str):
        is_visible = (
            ENV.from_string(is_visible).render(answers).strip() == "True"
        )

    if not cast(bool, is_visible):
        return False

    if "depends_on" not in question:
        return True

    depends_satisfied = True
    depends_on_attr = question.get("depends_on", [])

    if not isinstance(depends_on_attr, list):
        raise Exception("`depends_on` attribute is a list of dictionaries.")

    for criteria_or in depends_on_attr:
        if not isinstance(criteria_or, dict):
            raise Exception(
                "`depends_on` attribute is a list of dictionaries."
            )

        tmp_satisfied = False

        for crit_key, crit_value in criteria_or.items():
            if answers[crit_key] == crit_value:
                tmp_satisfied = True

        depends_satisfied = depends_satisfied and tmp_satisfied

    return depends_satisfied


def sanitize_package_slug(package_slug: str) -> str:
    """Filter to sanitize the package slug."""
    return re.sub(
        r"^\s+|\s+$", "", re.sub(r"[^a-zA-Z0-9_]+", "", package_slug)
    )


def make_questions(questions: dict[str, Any]) -> dict[str, str]:
    """Generate all the visible questions."""
    answers: dict[str, str] = {}

    # Get the size of the terminal window
    try:
        columns, _ = os.get_terminal_size()
    except OSError:
        columns = 80

    # Print a line
    print("-" * columns)
    print(
        Fore.YELLOW
        + "NOTE: "
        + Fore.RESET
        + "For questions with single or multiple choices, use the ARROW "
        + "keys to navigate through the options. Use the SPACE BAR to "
        + "select or deselect an option. Confirm your selection with "
        + "the ENTER key."
    )
    print("." * columns)

    # Create a Jinja2 environment and add the custom filter
    ENV.filters["sanitize_package_slug"] = sanitize_package_slug

    for question_id, question in questions.items():
        question_obj = _create_question(question_id, question)

        default_answer = question.get("default", "")
        default_answer = (
            ENV.from_string(default_answer).render(answers).strip()
        )

        # note: if question_object is None, that means that the question is
        #       not visible
        if not (check_visibility(question, answers) and question_obj):
            answers[question_id] = default_answer
            continue

        message = question.get("message", "")
        print(
            Style.BRIGHT
            + f"{message} ("
            + Fore.YELLOW
            + f"default: {default_answer}"
            + Fore.RESET
            + Style.BRIGHT
            + "):"
            + Fore.RESET
        )
        print(Fore.BLUE + ">> HELP: " + question["help"] + Fore.RESET)
        answer = inquirer.prompt([question_obj])

        # note: if answer is none, it means that the user cancelled
        #       the process.
        if answer is None:
            return {}
        answers[question_id] = answer.get(question_id, "") or default_answer
        print("." * columns)
    return answers



"""{{ cookiecutter.project_name }}."""
{%- if cookiecutter.use_black == "yes" %}
  {%- set QUOTE = '"' -%}
{%- else %}
  {%- set QUOTE = "'" -%}
{%- endif %}
{% if cookiecutter.build_system in ["hatch", "pdm", "flit"] -%}
__version__ = {{ QUOTE }}{{ cookiecutter.project_version }}{{ QUOTE }}
{%- else %}
from importlib import metadata as importlib_metadata


def get_version() -> str:
    """Return the program version."""
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return {{ QUOTE }}{{ cookiecutter.project_version }}{{ QUOTE }}  # semantic-release


version = get_version()

__version__ = version
{%- endif %}
__author__ = {{ QUOTE }}{{ cookiecutter.author_full_name }}{{ QUOTE }}
__email__ = {{ QUOTE }}{{ cookiecutter.author_email }}{{ QUOTE }}



"""Module with CLI functions."""

{%- if cookiecutter.command_line_interface == "Argparse" %}

import argparse

from {{ cookiecutter.package_slug }} import __version__


class CustomHelpFormatter(argparse.RawTextHelpFormatter):
    """Formatter for generating usage messages and argument help strings.

    Only the name of this class is considered a public API. All the methods
    provided by the class are considered an implementation detail.
    """

    def __init__(
        self,
        prog,
        indent_increment=2,
        max_help_position=4,
        width=None,
        **kwargs,
    ):
        super().__init__(
            prog,
            indent_increment=indent_increment,
            max_help_position=max_help_position,
            width=width,
            **kwargs,
        )


def get_args():
    """Return the arguments for the CLI."""
    parser = argparse.ArgumentParser(
        prog="{{ cookiecutter.project_slug }}",
        description=("{{ cookiecutter.project_name }}"),
        epilog=(
            "If you have any problem, open an issue at: "
            "{{ cookiecutter.project_url }}"
        ),
        add_help=True,
        formatter_class=CustomHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Show the version of the installed {{ cookiecutter.project_name }} tool.",
    )

    return parser


def show_version():
    """Show the version for the application."""
    print(__version__)


def app():
    """Run the application."""
    args_parser = get_args()
    args = args_parser.parse_args()

    if args.version:
        return show_version()

{%- elif cookiecutter.command_line_interface == "Click" %}

import click

from {{ cookiecutter.package_slug }} import __version__


@click.command()
@click.option(
    "--version",
    is_flag=True,
    help="Show the version of the installed {{ cookiecutter.project_name }} tool.",
)
def app(version):
    """Run the application."""
    if version:
        return click.echo(__version__)
    click.echo("You can add more Click commands here.")

{%- endif %}


if __name__ == "__main__":
    app()



"""Module with CLI functions."""

from __future__ import annotations

import argparse
import json
import os
import platform
import sys

from pathlib import Path
from typing import Union

from cookiecutter.main import cookiecutter

from scicookie.logs import SciCookieErrorType, SciCookieLogs
from scicookie.profile import Profile
from scicookie.ui import make_questions

PACKAGE_PATH = Path(__file__).parent
COOKIECUTTER_FILE_PATH = PACKAGE_PATH / "cookiecutter.json"


class CustomHelpFormatter(argparse.RawTextHelpFormatter):
    """Formatter for generating usage messages and argument help strings.

    Only the name of this class is considered a public API. All the methods
    provided by the class are considered an implementation detail.
    """

    def __init__(
        self,
        prog,
        indent_increment=2,
        max_help_position=4,
        width=None,
        **kwargs,
    ):
        """Define the parameters for the argparse help text."""
        super().__init__(
            prog,
            indent_increment=indent_increment,
            max_help_position=max_help_position,
            width=width,
            **kwargs,
        )


def _get_cookiecutter_default_answer(
    answer_definition: Union[str, list],
) -> str:
    if not isinstance(answer_definition, (str, list)):
        SciCookieLogs.raise_error(
            "Invalid cookiecutter configuration.",
            SciCookieErrorType.SCICOOKIE_INVALID_CONFIGURATION,
        )

    if isinstance(answer_definition, str):
        return answer_definition

    return answer_definition[0]


def call_cookiecutter(profile: Profile, answers: dict) -> None:
    """Call cookiecutter/cookiecutter with the parameters from the TUI."""
    answers_profile = {}
    questions = profile.config

    with open(COOKIECUTTER_FILE_PATH) as f:
        cookiecutter_config = json.load(f)

    # fill the answers with default value
    for question_id, question in questions.items():
        if question.get("control_flow", False):
            # "control_flow" is not defined in cookiecutter config
            continue

        if not question.get("visible", False):
            # get the default for non visible questions
            answers_profile[question_id] = question.get("default")
            continue

        if question.get("type") == "multiple-choices":
            for choice in question.get("choices", {}):
                choice_id = f"use_{choice.replace('-', '_')}"
                answers_profile[choice_id] = "no"
            continue

        answers_profile[question_id] = question.get(
            "default"
        ) or _get_cookiecutter_default_answer(cookiecutter_config[question_id])

    for question_id, answer in answers.items():
        if answer in [None, ""] or questions[question_id].get("control_flow"):
            continue

        if questions[question_id].get("type") != "multiple-choices":
            answers_profile[question_id] = answer
            continue

        answer_clean: list[str] = (
            answer if isinstance(answer, list) else [answer]
        )

        for choice in answer_clean:
            choice_id = f"use_{choice.replace('-', '_')}"
            answers_profile[choice_id] = "yes"

    env_path_sep = ";" if platform.system() == "Windows" else ":"
    pipx_path: list[str] = list(filter(lambda v: ".local/pipx" in v, sys.path))
    new_path: list[str] = [*pipx_path, os.getenv("PATH", "")]
    os.environ["PATH"] = env_path_sep.join(new_path)

    # TODO: print answers with --verbose
    # print(answers_profile)

    cookiecutter(
        str(PACKAGE_PATH), no_input=True, extra_context=answers_profile
    )


def call_app(profile_name: str) -> None:
    """Call the scicookie app."""
    profile = Profile(profile_name)

    answers = make_questions(profile.config)

    if not answers:
        return

    call_cookiecutter(profile, answers)


def app() -> None:
    """Run SciCookie."""
    # note: this parameter should be provided by a CLI argument

    parser = argparse.ArgumentParser(
        prog="SciCookie",
        description=(
            "SciCookie is a template developed by Open Science Labs that "
            "creates projects through different options of profile. "
            "It serves as a boilerplate which can be used by beginners as "
            "well as full fledged developers to simplify the project creation "
            "process and save considerable amount of time. It creates  "
            "projects with an initial layout that includes recommended  "
            "tools, workflows, and project structure."
        ),
        epilog=(
            "If you have any problem, open an issue at: "
            "https://github.com/osl-incubator/scicookie"
        ),
        add_help=True,
        formatter_class=CustomHelpFormatter,
    )
    parser.add_argument(
        "--profile",
        type=str,
        default="base",
        help="Select the profile to be used",
    )

    args = parser.parse_args()
    call_app(args.profile)



"""

Generate the code reference pages and navigation.

REF:
https://github.com/mkdocstrings/mkdocstrings/blob/main/scripts/gen_ref_nav.py

"""

{%- if cookiecutter.use_black == "yes" %}
  {%- set QUOTE = '"' -%}
{%- else %}
  {%- set QUOTE = "'" -%}
{%- endif %}

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()
mod_symbol = (
    '<code class="doc-symbol doc-symbol-nav doc-symbol-module"></code>'
)

root = Path(__file__).parent.parent
{% if cookiecutter.project_layout == 'src' -%}
src = root / {{ QUOTE }}src{{ QUOTE }}
{% else -%}
src = root
{% endif -%}

for path in sorted(src.rglob({{ QUOTE }}*.py{{ QUOTE }})):
    module_path = path.relative_to(src).with_suffix({{ QUOTE }}{{ QUOTE }})
    doc_path = path.relative_to(src / {{ QUOTE }}{{ cookiecutter.package_slug }}{{ QUOTE }}).with_suffix({{ QUOTE }}.md{{ QUOTE }})
    full_doc_path = Path({{ QUOTE }}api{{ QUOTE }}, doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == {{ QUOTE }}__init__{{ QUOTE }}:
        parts = parts[:-1]
        doc_path = doc_path.with_name({{ QUOTE }}index.md{{ QUOTE }})
        full_doc_path = full_doc_path.with_name({{ QUOTE }}index.md{{ QUOTE }})
    elif parts[-1].startswith({{ QUOTE }}_{{ QUOTE }}):
        continue

    nav_parts = [f{{ QUOTE }}{mod_symbol} {part}{{ QUOTE }} for part in parts]
    nav[tuple(nav_parts)] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, {{ QUOTE }}w{{ QUOTE }}) as fd:
        ident = {{ QUOTE }}.{{ QUOTE }}.join(parts)
        fd.write(f{{ QUOTE }}::: {ident}{{ QUOTE }})

    mkdocs_gen_files.set_edit_path(
        full_doc_path, {{ QUOTE }}..{{ QUOTE }} / path.relative_to(root)
    )

with mkdocs_gen_files.open({{ QUOTE }}api/SUMMARY.md{{ QUOTE }}, {{ QUOTE }}w{{ QUOTE }}) as nav_file:
    nav_file.writelines(nav.build_literate_nav())



"""
Entrypoint module, in case you use `python -m {{ cookiecutter.project_slug }}`.

Why does this file exist, and why __main__? For more info, read:
- https://www.python.org/dev/peps/pep-0338/
- https://docs.python.org/2/using/cmdline.html#cmdoption-m
- https://docs.python.org/3/using/cmdline.html#cmdoption-m
"""
{%- if cookiecutter.use_black == "yes" %}
  {%- set QUOTE = '"' -%}
{%- else %}
  {%- set QUOTE = "'" -%}
{%- endif %}

from {{cookiecutter.package_slug}}.cli import main

if __name__ == {{ QUOTE }}__main__{{ QUOTE }}:
    main()



from setuptools import setup  # isort:skip

# Available at setup time due to pyproject.toml
from pybind11.setup_helpers import Pybind11Extension  # isort:skip

# Note:
#   Sort input source files if you glob sources to ensure bit-for-bit
#   reproducible builds (https://github.com/pybind/python_example/pull/53)

ext_modules = [
    Pybind11Extension(
        "{{ cookiecutter.project_slug }}._core",
        ["src/main.cpp"],
        cxx_std=11,
    ),
]


setup(
    ext_modules=ext_modules,
)



"""Tool for fixing the end of file."""

from pathlib import Path

from pre_commit_hooks import end_of_file_fixer


def is_binary(file_path) -> bool:
    """Check if the given file is binary."""
    try:
        with open(file_path, "rb") as file:
            textchars = bytearray(
                {7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F}
            )
            return bool(file.read(1024).translate(None, textchars))
    except Exception as e:
        print(f"Error opening file {file_path}: {e}")
        return True


def run(dir_path: Path) -> None:
    """Fix the end of file for all files recursively inside a directory."""
    base_path = dir_path.absolute()
    files = []
    for file_path in base_path.rglob("*"):
        if (
            ".git" not in file_path.parts
            and file_path.is_file()
            and not is_binary(file_path)
        ):
            files.append(str(file_path))
    end_of_file_fixer.main(files)



"""Profile handles "profiles" defined in the .yaml files."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml

from scicookie.logs import SciCookieErrorType, SciCookieLogs

PACKAGE_PATH = Path(__file__).parent
PROFILE_DIR_PATH = PACKAGE_PATH / "profiles"


class Profile:
    """Profile class that handles profiles defined in the .yaml files."""

    profile_name: str = ""
    config: dict[str, Any] = {}  # noqa: RUF012
    profiles_available: list[str] = []  # noqa: RUF012

    def __init__(self, profile_name: str):
        self._load_profiles_available()

        if profile_name not in self.profiles_available:
            SciCookieLogs.raise_error(
                f"The given profile ({profile_name}) is not available.",
                SciCookieErrorType.SCICOOKIE_INVALID_PARAMETER,
            )

        self.profile_name = profile_name
        self.config = self.read_config()

    def _load_profiles_available(self):
        self.profiles_available = []

        profiles_path = Path(__file__).absolute().parent / "profiles"

        for file in profiles_path.glob("*.yaml"):
            self.profiles_available.append(file.stem)

    def read_config(self):
        """Read the config file."""
        config = {}
        with open(PROFILE_DIR_PATH / "base.yaml") as f:
            config = yaml.safe_load(f)

        with open(PROFILE_DIR_PATH / f"{self.profile_name}.yaml") as f:
            config_profile = yaml.safe_load(f)
            for name, properties in config_profile.items():
                config[name].update(properties)

        return config
